"""Report component for displaying information from wf-clone-validation."""
import argparse

from bokeh.layouts import gridplot, layout
from bokeh.models import Div, Panel, Tabs
import markdown
import pandas as pd

from aplanat import base, hist
from aplanat.report import bokeh_table, HTMLReport, HTMLSection
from aplanat.util import Colors


class CloneValidationQC(HTMLSection):
    """Build an aplanat dashboard from wf-clone-validation output."""

    def __init__(
        self,
        assembly_summary,
        assembly_mafs,
        reads_summary,
        fastq_summary
    ) -> None:
        """Load the json file and output the dashboard."""
        super().__init__()
        self.md = markdown.Markdown()
        self.build_report(
            assembly_summary,
            assembly_mafs,
            reads_summary,
            fastq_summary
        )

    def build_report(
        self,
        assembly_summary,
        assembly_mafs,
        reads_summary,
        fastq_summary
    ):
        """Build_report."""
        assemblies = self.build_assemblies_panel(
            assembly_summary, assembly_mafs)
        samples = self.build_samples_panel(
            fastq_summary, reads_summary)

        # Tabula rasa
        panel = Tabs(tabs=[assemblies, samples])
        self.plot(panel)

    def build_assemblies_panel(self, assemblies_file, mafs):
        """Build_assemblies_panel."""
        # Dotplots
        dotplots = []
        for maf in mafs:
            dotplots.append(self.dotplot_assembly(maf))

        text = (
            "This tab contains a basic table of "
            "stats for each final assembly, as well "
            "as dotplots generated by self-alignment. "
            "These dotplots should be a near-perfect "
            "diagonal line if the assembly has worked."
        )
        plots = [
            [self.get_description(text)],
            [Div(text="<br />")],
            [self.build_table(assemblies_file)],
            [gridplot(dotplots, ncols=2)]
        ]

        main = layout(plots, sizing_mode="scale_width")
        panel = Panel(child=main, title="Assemblies")
        return panel

    def build_samples_panel(self, summary_file, reads_file):
        """Build_length_tab."""
        summary_df = pd.read_csv(summary_file, sep="\t")
        reads_df = pd.read_csv(reads_file, sep="\t")

        length_plots = []
        qual_plots = []

        filenames = set(reads_df['filename'].tolist())
        for fname in filenames:
            fname_read = reads_df.loc[reads_df['filename'] == fname]
            fname_summary = summary_df.loc[summary_df['filename'] == fname]
            length_plots.append(self.plot_read_length_distribution(
                fname, int(fname_summary['n_seqs']),
                int(fname_summary['n_bases']),
                int(fname_summary['min_length']),
                int(fname_summary['max_length']), fname_read))
            qual_plots.append(self.plot_qscore_distribution(
                fname, float(fname_summary['mean_quality']), fname_read))

        length_grid = gridplot(
            length_plots, ncols=2,
            sizing_mode="stretch_width")

        quality_grid = gridplot(
            qual_plots, ncols=2,
            sizing_mode="stretch_width")

        text = (
            "This tab contains visualisations of "
            "read length and average quality for "
            "the final, filtered datasets used in assembly, "
            "from each of the samples respectively. I.e, "
            "this is post-host filtering and downsampling."
        )
        plots = [
            [self.get_description(text)],
            [length_grid],
            [quality_grid]
        ]

        main = layout(plots, sizing_mode="scale_width")
        return Panel(child=main, title="Samples")

    def build_table(self, tsv_file):
        """Build the table or get the flake again."""
        df = pd.read_csv(tsv_file, sep='\t')
        plot = bokeh_table(df, index=False)
        plot.height = min(plot.height, 25*(len(df) + 1))
        return plot

    def plot_read_length_distribution(
            self, fname, nseqs, nbases, minl, maxl, data):
        """Plot_read_length_distribution."""
        plot = hist.histogram(
            [data['read_length'].tolist()],
            # bins=1000,
            height=300,
            width=400,
            xlim=(0, max(data['read_length']) + 200),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Length',
            y_axis_label='Count',
            title=f"{fname}: {nseqs} seqs, {nbases} bp, {minl} min, {maxl} max"
        )

        self.style_plot(plot)
        return plot

    def plot_qscore_distribution(self, fname, mean, data):
        """Plot_qscore_distribution."""
        plot = hist.histogram(
            [data['mean_quality'].tolist()],
            # bins=600,
            height=300,
            width=400,
            xlim=(0, 30),
            colors=[Colors.light_cornflower_blue],
            x_axis_label='Mean Quality',
            y_axis_label='Count',
            title=f"{fname}: {mean} mean q-score"
        )

        self.style_plot(plot)
        return plot

    def style_plot(self, plot):
        """Style plots."""
        plot.margin = (10, 10, 40, 10)
        plot.background_fill_alpha = 0

    def dotplot_assembly(self, assembly_maf):
        """Dotplot the assembly."""
        records = list()
        with open(assembly_maf) as maf:
            while True:
                line = maf.readline()
                print(line)
                if line.startswith('#'):
                    continue
                elif line.startswith('a'):
                    r1 = maf.readline().split()[1:5]
                    r2 = maf.readline().split()[1:5]
                    maf.readline()
                    records.append(r1 + r2)
                elif line == "":
                    break
                else:
                    print(line)
                    raise IOError("Cannot read alignment file")

        names = ['ref', 'rstart', 'rlen', 'rorient',
                 'query', 'qstart', 'qlen', 'qorient']
        df = pd.DataFrame(records, columns=names)
        df = df.loc[df['qorient'] == '+']
        for f in ['qstart', 'qlen', 'rstart', 'rlen']:
            df[f] = df[f].astype(int)
        df['qend'] = df['qstart'] + df['qlen']
        df['rend'] = df['rstart'] + df['rlen']

        df['qend'] = df['qend'].astype(int)
        df['rend'] = df['rend'].astype(int)
        dotplot = base.simple(
            [], [],
            xlim=(0, max(df['rend'])),
            ylim=(0, max(df['qend'])),
            width=440,
            height=400,
            x_axis_label='position',
            y_axis_label='position',
            title=assembly_maf
        )
        dotplot.segment(df['rstart'], df['qstart'], df['rend'], df['qend'])
        return dotplot

    def get_description(self, desc):
        """Get Description."""
        styles = [
            "display:block;",
            "width:100%;",
            "padding:25px 0 0 0;",
            "font-size: 16px;",
            "margin-bottom: 0;"
        ]
        return Div(
            text=f'<p style="{"".join(styles)}" class="lead">{desc}</p>'
        )


def main(args):
    """Entry point to create a wf-clone-validation report."""
    report = HTMLReport(
        title='Clone Validaton',
        lead=(
            "Results generated through the wf-clone-validation "
            "nextflow workflow provided by Oxford Nanopore "
            "Technologies"))
    report.add_section(
        section=CloneValidationQC(
            assembly_summary=args.assembly_summary,
            assembly_mafs=args.assembly_mafs,
            reads_summary=args.reads_summary,
            fastq_summary=args.fastq_summary
        ))
    report.write('report.html')


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Clone Validation QC report',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)

    parser.add_argument(
        "--assembly_summary",
        required=True
    )

    parser.add_argument(
        "--assembly_mafs",
        nargs='*',
        required=True
    )

    parser.add_argument(
        "--reads_summary",
        required=True
    )

    parser.add_argument(
        "--fastq_summary",
        required=True
    )

    return parser



"""Utility functions for aiding plotting."""

from bokeh import palettes
import numpy as np
import pandas as pd
from scipy import stats as sp_stats


def read_files(summaries):
    """Read a set of files and join to single dataframe."""
    dfs = list()
    if not isinstance(summaries, (list, tuple)):
        summaries = [summaries]
    for fname in sorted(summaries):
        dfs.append(pd.read_csv(fname, sep="\t"))
    return pd.concat(dfs)


class _colors:
    """Some colours that someone thought were nice."""

    cerulean = "#0084A9"
    not_black = "#001A21"
    feldgrau = "#455556"
    dim_gray = "#666666"
    light_cornflower_blue = "#90C5E7"
    dark_gray = "#B5AEA7"
    isabelline = "#F0EFED"
    medium_spring_bud = "#B8E986"
    cinnabar = "#EF4134"
    sandstorm = "#F5CC49"
    fandango = "#A53F96"
    green = "#17BB75"
    verdigris = "#54B8B1"


Colors = _colors()


class Limiter(object):
    """Helper class to determine limits of data."""

    def __init__(self, limits=None):
        """Initialize the limit gathering."""
        if limits is not None:
            self.min, self.max = limits
        else:
            self.min, self.max = +float('inf'), -float('inf')

    def accumulate(self, data):
        """Analyse data to update limits.

        :param data: new data to analyse.

        """
        self.min = min(self.min, min(data))
        self.max = max(self.max, max(data))
        return self

    def fix(self, lower=None, upper=None):
        """Fix limits to new values.

        :param lower: new lower limit.
        :param upper: new upper limit.

        """
        if lower is not None:
            self.min = lower
        if upper is not None:
            self.max = upper
        return self

    def __repr__(self):
        """Return string representation of self."""
        return repr((self.min, self.max))


def pad(data):
    """Attempt to find a sensible plot bounds for a vector."""
    uniq = sorted(np.unique(data))
    pad = 0.5 * min(np.ediff1d(uniq))
    return uniq[0] - pad, uniq[-1] + pad


def kernel_density_estimate(x, step=0.2):
    """Kernel density to approximate distribution.

    :param x: data of which to find mode.
    :param step: discretization of KDE PDF.
    """
    # estimate bandwidth of kde, R's nrd0 rule-of-thumb
    hi = np.std(x, ddof=1)
    q75, q25 = np.percentile(x, [75, 25])
    iqr = q75 - q25
    lo = min(hi, iqr/1.34)
    if not ((lo == hi) or (lo == abs(x[0])) or (lo == 1)):
        lo = 1
    bw = 0.9 * lo * len(x)**-0.2

    # create a KDE
    x_grid = np.arange(min(x), max(x), step)
    kernel = sp_stats.gaussian_kde(x, bw_method=bw)
    pdf = kernel(x_grid)
    return x_grid, pdf


def choose_palette(ncolours):
    """Choose colour palette.

    :param: ncolours: number of colours.
    """
    if ncolours <= 20:
        # Category20 contains dark/light pairs, avoid using these
        pal = palettes.Category20[20]
        palette = pal[0::2] + pal[1::2]
        cols = palette[0:ncolours]
    elif ncolours <= 256:
        cols = palettes.Viridis256[::256//ncolours]
    else:
        raise ValueError(
            "Cannot create colour palette with more than 256 colours.")
    return cols



"""Report building from multiple items."""

from collections import OrderedDict
import uuid

from bokeh.embed import components
from bokeh.model import Model
from bokeh.models.sources import ColumnDataSource
from bokeh.models.widgets import DataTable, TableColumn
from bokeh.resources import INLINE
from jinja2 import Template
import markdown


def _maybe_new_report(section, require_keys=False):
    """Create a new report section, or returns input."""
    if section is None:
        section = HTMLSection(require_keys=require_keys)
    else:
        if not isinstance(section, HTMLSection):
            raise TypeError("`section` should be an `HTMLSection`.")
    return section


class HTMLSection(OrderedDict):
    """A section of a report.

    The class can be extended o provide novel report components by
    overriding `._extra_components()`.
    """

    def __init__(self, require_keys=False):
        """Initialize the report item collection.

        :param title: report title.
        :param lead: report strapline, shown below title.
        :param require_keys: require keys when adding items.
        """
        self.require_keys = require_keys
        self.plots = list()
        self.md = markdown.Markdown()

    def _add_item(self, item, key=None):
        """Add an item to the report.

        :param item: item to add.
        :param key: unique key for item.
        """
        if key is None:
            if self.require_keys:
                raise ValueError('A key is required.')
            else:
                key = str(uuid.uuid4())
        self[key] = item

    def placeholder(self, key):
        """Add a placeholder to be filled in later."""
        self._add_item(None, key=key)

    def plot(self, plot, key=None):
        """Add a plot to the report.

        :param plot: bokeh plot instance.
        :param key: unique key for item.
        """
        self._add_item(plot, key=key)

    def table(self, df, index=True, key=None, shrink=True, **kwargs):
        """Add a pandas dataframe to the report.

        :param df: pandas dataframe instance.
        :param index: include dataframe index in output.
        :param key: unique key for item.
        :param shrink: shrink wrap the table to avoid whitespace.
        :param kwargs: passed to bokeh DataTable.
        """
        plot = bokeh_table(df, index=index, **kwargs)
        plot.height = min(plot.height, 25*(len(df) + 1))
        self.plot(plot, key)

    def markdown(self, text, key=None):
        """Add markdown formatted text to the report.

        :param text: markdown formatted text.
        :param key: unique key for item.
        """
        if text is None or text == '':
            return
        html = self.md.convert(text)
        self.md.reset()
        self._add_item(html, key=key)

    def _plot_components(self):
        """Return html script and div tags for bokeh plots."""
        # handle bokeh plots
        scripts = list()
        divs = dict()
        plots = {k: v for k, v in self.items() if isinstance(v, Model)}
        if len(plots) > 0:
            scripts, divs = components(plots)
            scripts = [scripts]
        return scripts, divs

    def _table_components(self):
        """Return html script and div for tables."""
        # TODO: reimplement `.table()` to not use bokeh?
        pass

    def _extra_components(self):
        """Return additional script and divs."""
        # hook for derived classes:
        # return: list of scripts, dict of divs keyed on self.keys
        return list(), dict()

    def components(self):
        """Fetch script and div tags for report."""
        scripts = list()
        specials = dict()
        # retrieve bokeh plot divs
        plot_scripts, plot_divs = self._plot_components()
        scripts.extend(plot_scripts)
        specials.update(plot_divs)
        # TODO: add special handing for new tables?
        # retrieve any extra specials
        extra_scripts, extra_divs = self._extra_components()
        scripts.extend(extra_scripts)
        specials.update(extra_divs)

        # put things together in order
        divs = list()
        for k in self.keys():
            try:
                divs.append(specials[k])
            except KeyError:
                if self[k] is None:
                    raise ValueError(
                        "Placeholder `{}` was not assigned "
                        "a value.".format(k)
                        )
                # fall back to putting items in directly
                divs.append(self[k])
        return scripts, divs


class HTMLReport(HTMLSection):
    """Generate HTML Report from a series of bokeh figures.

    Items added to the report take an optional key argument, adding items
    with the same key allows an update in place whilst maintaining the order
    in which items were added. Items can be grouped into sections for easier
    out of order addition.
    """

    def __init__(self, title="", lead="", require_keys=False):
        """Initialize the report item collection.

        :param title: report title.
        :param lead: report strapline, shown below title.
        :param require_keys: require keys when adding items.
        """
        super().__init__(require_keys=require_keys)
        self.title = title
        self.lead = lead
        self.sections = OrderedDict()
        self.sections['main'] = self
        self.template = Template(
            """\
            <!doctype html>
            <html lang="en">
            <head>
                <meta charset="utf-8">
                <meta http-equiv="X-UA-Compatible" content="IE=edge">
                <meta name="viewport" content="width=device-width, initial-scale=1">
                    <title>{{ title }}</title>
<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/3.4.1/css/bootstrap.min.css" integrity="sha384-HSMxcRTRxnN+Bdg0JdbxYKrThecOKuH5zCYotlSAcp1+c8xmyTe9GYg1l9a69psu" crossorigin="anonymous">

                    {{ resources }}
                    {{ script }}
                </head>
                <body>
            <div class="container">
              <h1>{{ title }}</h1>
              <p class="lead">{{ lead }}
            {{ div }}
                </body>
            </html>
            """  # noqa
        )

    def add_section(self, key=None, section=None, require_keys=False):
        """Add a section (grouping of items) to the report.

        :param key: unique key for section.
        :param section: `HTMLSection` to add rather than creating anew.

        :returns: the report section.

        """
        if key is None:
            key = str(uuid.uuid4())
        section = _maybe_new_report(section, require_keys=require_keys)
        self.sections[key] = section
        return self.sections[key]

    def render(self):
        """Generate HTML report containing figures."""
        resources = INLINE.render()

        all_scripts = list()
        all_divs = list()
        for sec_name, section in self.sections.items():
            scripts, divs = section.components()
            all_scripts.extend(scripts)
            all_divs.extend(divs)

        script = '\n'.join(all_scripts)
        divs = '\n'.join(all_divs)
        return self.template.render(
            title=self.title, lead=self.lead,
            resources=resources, script=script, div=divs)

    def write(self, path):
        """Write html report to file."""
        with open(path, "w", encoding='utf8') as outfile:
            outfile.write(self.render())


def bokeh_table(df, index=True, **kwargs):
    """Generate a bokeh table from a pandas dataframe."""
    columns = [TableColumn(field=x, title=x) for x in df.columns]
    if index:
        columns = [TableColumn(field="index", title="Feature")] + columns
    return DataTable(
            columns=columns, source=ColumnDataSource(df), index_position=None,
            **kwargs)



"""Bio specific plots, could be generalised."""

from io import StringIO

from bokeh.models import Range1d
from bokeh.models.tickers import FixedTicker
from bokeh.plotting import figure
import numpy as np
import pandas as pd

_chrom_data_ = """
chrom length
1 249250621
2 243199373
3 198022430
4 191154276
5 180915260
6 171115067
7 159138663
8 146364022
9 141213431
10 135534747
11 135006516
12 133851895
13 115169878
14 107349540
15 102531392
16 90354753
17 81195210
18 78077248
19 59128983
20 63025520
21 48129895
22 51304566
X 155270560
Y 59373566
"""
_chrom_data_ = pd.read_csv(StringIO(_chrom_data_), delimiter=" ")


def karyotype(
        x_datas, y_datas, names=None, colors=None, alpha=0.2,
        chrom_data=_chrom_data_, **kwargs):
    """Create a heatmap from three columns.

    :param x: x-axis coordinates (chromosome position)
    :param y: y-axis coordinates (chromosome)
    :param name: title for z-axis.
    :param colors: colours to plot tics for each data series.
    :param alpha: alpha level (transparency) with which to plot tics.
    :params chrom_data: a dataframe (or ndarray) with fields
        `chrom`, and `length` used to plot chromosome outline
        rectangles. Chromosomes will be plotted vertically (top to bottom)
        in the order given by this input.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    if names is None:
        names = [None] * len(x_datas)
    if colors is None:
        colors = [None] * len(x_datas)
    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)
    p = figure(**defaults)
    width = 0.8
    chrom_order = pd.Series(range(1, len(chrom_data) + 1))
    p.rect(
        chrom_data["length"] // 2, chrom_order,
        chrom_data["length"], width,
        fill_color='white', line_color='black')

    plot_order = dict(zip(chrom_data['chrom'], chrom_order))
    for x, y, name, color in zip(x_datas, y_datas, names, colors):
        y = np.array([plot_order[i] for i in y])
        kw = {'alpha': alpha}
        if name is not None:
            kw['legend_label'] = name
        if color is not None:
            kw['color'] = color
        p.segment(
            x, y - width / 2, x, y + width / 2, **kw)

    # set up the axes
    p.xaxis.formatter.use_scientific = False
    p.yaxis.ticker = FixedTicker()
    p.yaxis.ticker.ticks = chrom_order
    p.yaxis.major_label_overrides = {v: k for k, v in plot_order.items()}
    p.x_range = Range1d(
        start=0, end=max(chrom_data['length']),
        bounds=(0, max(chrom_data['length'])))
    miny, maxy = 0, len(chrom_data) + 1
    p.y_range = Range1d(
        start=miny, end=maxy, bounds=(miny, maxy))
    return p



"""Simple bokeh plotting API."""

import argparse
import importlib
import warnings

from bokeh.colors import RGB
import bokeh.io as bkio
from bokeh.layouts import gridplot
from bokeh.plotting import Figure

__version__ = "0.3.9"

# we don't run a comprehensive test suite and mostly in notebooks,
# so show warnings all the time.
warnings.simplefilter('always', DeprecationWarning)


class Grid(list):
    """Create a list of lists of fixed length."""

    def __init__(self, width=4):
        """Initialize the Grid."""
        self.width = width
        self.append(list())

    def add(self, item):
        """Add an item to the grid.

        :param item: item to add

        """
        self[-1].append(item)
        if len(self[-1]) == self.width:
            self.append(list())

    def extend(self, items):
        """Add multiple items.

        :param items: items to add.

        """
        for item in items:
            self.add(item)


def all_children(plotlike):
    """Recursively find all Figures in a plotting layout.

    :param plotlike: a bokeh layout or plot.
    """
    my_children = []
    if hasattr(plotlike, 'children'):
        for child in plotlike.children:
            if isinstance(child, tuple):
                child = child[0]
            my_children.extend(all_children(child))
    if isinstance(plotlike, Figure):
        my_children.append(plotlike)
    return my_children


def show(plot, background=None):
    """Show a plot in a notebook.

    :param background: a fill colour for the plot background, hex string or RGB
        tuple.

    """
    children = all_children(plot)
    orig_colours = None
    if background is not None:
        if isinstance(background, tuple):
            background = RGB(*background)
        elif not (isinstance(background, str) and background.startswith('#')):
            raise TypeError(
                "`background` should be a RGB tuple or hex-colour string.")
        orig_colours = list()
        for child in children:
            orig_colours.append((
                child.background_fill_color,
                child.border_fill_color))
            child.background_fill_color = background
            child.border_fill_color = background
    bkio.output_notebook(hide_banner=True)
    bkio.show(plot)
    if background is not None:
        for colours, child in zip(orig_colours, children):
            child.background_fill_color = colours[0]
            child.border_fill_color = colours[1]


def grid(plots, ncol=4, display=True, **kwargs):
    """Show a grid of plots in a notebook.

    :param plots: a list of bokeh plots.
    :param ncol: number of columns.
    :param kwargs: kwargs for bokeh `gridplot`.

    """
    warnings.warn(
        "Please use `bokeh.layout.gridplot` directly with an optional "
        "call to `.show()`", DeprecationWarning)

    grid = Grid(width=ncol)
    grid.extend(plots)
    plot = gridplot(grid, **kwargs)
    if display:
        show(plot)
    else:
        return plot


def InfoGraphItems():
    """Cumulatively create items for an infographic."""
    from aplanat import graphics
    warnings.warn(
        "This class has been moved to `aplanat.graphics.InfoGraphItems()`",
        DeprecationWarning)
    return graphics.InfoGraphItems()


def infographic(items, **kwargs):
    """Create an infographic 'plot'."""
    from aplanat import graphics
    warnings.warn(
        "This function has been moved to `aplanat.graphics.infographic()`",
        DeprecationWarning)
    return graphics.infographic(items, **kwargs)


def cli():
    """Run aplanat entry point."""
    parser = argparse.ArgumentParser(
        'aplanat',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        '-v', '--version', action='version',
        version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True

    # add reporting modules
    modules = ['bcfstats', 'mapula', 'nextclade', 'fastcat', 'clonevalidation']
    for module in modules:
        mod = importlib.import_module('aplanat.components.{}'.format(module))
        p = subparsers.add_parser(module, parents=[mod.argparser()])
        p.set_defaults(func=mod.main)

    args = parser.parse_args()
    args.func(args)



"""Report components for displaying information from bcftools stats."""

import argparse

from bokeh.models import LinearColorMapper
from bokeh.palettes import Blues9
from bokeh.plotting import figure
import pandas as pd

from aplanat import hist
from aplanat.parsers.bcfstats import parse_bcftools_stats_multi
from aplanat.report import _maybe_new_report, HTMLReport
from aplanat.util import Colors


def variant_counts_table(
        bcf_stats, samples_as_columns=False,
        header="**Variant counts:**", report=None):
    """Create a report section contains variant counts.

    :param bcf_stats: one or more outputs from `bcftools stats`.
    :param sample_as_columns: transpose table to put data for each
        sample into a column.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)

    df = bcf_stats['SN'].drop(columns='samples')
    index = False
    if samples_as_columns:
        df = df.set_index('sample').transpose()
        index = True
    report.table(df, index=index)
    return report


def trans_counts(
        bcf_stats, samples_as_columns=False,
        header="**Transitions and tranversions:**", report=None):
    """Create a report section with transition and transversion counts.

    :param bcf_stats: one or more outputs from `bcftools stats`.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)

    df = bcf_stats['TSTV']
    index = False
    if samples_as_columns:
        df = df.set_index('sample').transpose()
        index = True
    report.table(df, index=index)
    return report


_sub_header = """
**Substitution types**

Base substitutions aggregated across all samples (symmetrised by pairing)
"""


def sub_matrix(bcf_stats, header=_sub_header, report=None):
    """Create a report section with a base substitution matrix.

    :param bcf_stats: one or more outputs from `bcftools stats`.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)

    sim_sub = {
        'G>A': 'C>T', 'G>C': 'C>G', 'G>T': 'C>A',
        'T>A': 'A>T', 'T>C': 'A>G', 'T>G': 'A>C'}

    def canon_sub(sub):
        b1 = sub[0]
        if b1 not in {'A', 'C'}:
            return canon_sub(sim_sub[sub])
        else:
            return b1, sub[2]

    df = bcf_stats['ST']
    df['canon_sub'] = df['type'].apply(canon_sub)
    df['original'] = df['canon_sub'].apply(lambda x: x[0])
    df['substitution'] = df['canon_sub'].apply(lambda x: x[1])
    df['count'] = df['count'].astype(int)
    df = df[['original', 'substitution', 'count']] \
        .groupby(['original', 'substitution']) \
        .agg(count=pd.NamedAgg(column='count', aggfunc='sum')) \
        .reset_index()

    colors = Blues9[::-1]
    mapper = LinearColorMapper(
        palette=colors, low=min(df['count']), high=max(df['count']))
    p = figure(
        y_range=['C', 'A'], x_range=['A', 'C', 'G', 'T'],
        x_axis_location="above",
        x_axis_label='alternative base',
        y_axis_label='reference base',
        tools="save", toolbar_location='below',
        output_backend="webgl",
        height=225, width=300,
        tooltips=[('sub', '@original>@substitution'), ('count', '@count')])
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.rect(
        source=df, y="original", x="substitution", width=1, height=1,
        fill_color={'field': 'count', 'transform': mapper},
        line_color=None)
    report.plot(p)
    return report


_indel_length_header = """
**Indel lengths**

Insertion and deletion lengths aggregated across all samples.
"""


def indel_lengths(
        bcf_stats, header=_indel_length_header, report=None,
        color=Colors.light_cornflower_blue):
    """Create a report section containing an indel length chart.

    :param bcf_stats: one or more outputs from `bcftools stats`.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.
    :param color: color of bars in chart.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    report = _maybe_new_report(report)
    report.markdown(header)
    try:
        df = bcf_stats['IDD']
    except KeyError:
        # If there are no indels, bcftools doesn't contain the table
        report.markdown("*No indels to report.*")
    else:
        df['nlength'] = df['length (deletions negative)'].astype(int)
        df['count'] = df['number of sites'].astype(int)
        # pad just to pull out axes by a minimum
        pad = pd.DataFrame({'nlength': [-10, +10], 'count': [0, 0]})
        counts = df.groupby('nlength') \
            .agg(count=pd.NamedAgg(column='count', aggfunc='sum')) \
            .reset_index().append(pad)
        plot = hist.histogram(
            [counts['nlength']], weights=[counts['count']],
            colors=[color], binwidth=1,
            title='Insertion and deletion variant lengths',
            x_axis_label='Length / bases (deletions negative)',
            y_axis_label='Count')
        report.plot(plot)
    return report


_full_report_header = """
### Variant call summaries

The following tables and figures are derived from
the output of `bcftools stats`.
"""


def full_report(bcf_stats, header=_full_report_header, report=None):
    """Create a report section from the output of bcftools stats.

    :param bcf_stats: one or more outputs from `bcftools stats`.
    :param header: a markdown formatted header.
    :param report: an HTMLSection instance.

    :returns: an HTMLSection instance, if `report` was provided the given
        instance is modified and returned.
    """
    if not isinstance(bcf_stats, (list, tuple)):
        bcf_stats = [bcf_stats]
    bcf_stats = parse_bcftools_stats_multi(bcf_stats)

    report = _maybe_new_report(report)
    report.markdown(header)
    report = variant_counts_table(bcf_stats, report=report)
    report = trans_counts(bcf_stats, report=report)
    report = sub_matrix(bcf_stats, report=report)
    report = indel_lengths(bcf_stats, report=report)
    return report


def main(args):
    """Entry point to create a report from bcftools stats."""
    report = full_report(args.bcfstats_files, report=HTMLReport())
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'bcftools stats report',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--output", default="bcf_stats_report.html",
        help="Output HTML file.")
    parser.add_argument(
        "bcfstats_files", nargs='+',
        help="One or more files containing output from `bcftools stats.`")
    return parser



"""Nextclade reporting section."""

import argparse

from jinja2 import Template
import pkg_resources

from aplanat.report import HTMLReport, HTMLSection


class NextClade(HTMLSection):
    """A nextclade report component."""

    def __init__(self, json, add_title=True):
        """Initialize a report component.

        :param json: json output file from nextclade CLI.
        """
        super().__init__()
        template = Template(
            """\
            <div>
            <nxt-table>
                <script defer="">
                const data = {{ data }}
                var nxt = document.querySelector('nxt-table')
                nxt.data = data
                </script>
            </nxt-table>
            </div>
            """  # noqa
        )
        if add_title:
            self.markdown('''
### NextClade analysis
The following view is produced by the
[nextclade](https://clades.nextstrain.org/) software.
''')
        with open(json, encoding='utf8') as fh:
            self._add_item(
                template.render(data=fh.read()))
        script = pkg_resources.resource_filename(
            'aplanat', 'data/nextclade.html')
        with open(script, encoding='utf8') as fh:
            self._script = fh.read()

    def _extra_components(self):
        """Return script and div for report."""
        # out main content div will be taken care of by the parent class
        return [self._script], dict()


def main(args):
    """Entry point to create a report from nextclade."""
    report = HTMLReport()
    report.add_section(section=NextClade(args.json))
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'nextclade report',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--output", default="nextclade_report.html",
        help="Output HTML file.")
    parser.add_argument(
        "json",
        help="JSON output file from nextclade CLI.")
    return parser
